/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements. See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.kafka.connect.mirror.utils;

import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.mockito.stubbing.OngoingStubbing;

import java.util.Optional;
import java.util.concurrent.Callable;

import static org.hamcrest.CoreMatchers.is;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.atMost;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

@ExtendWith(MockitoExtension.class)
public class RetrierTest {

    final String expectedResult = "result";


    @Mock(lenient = true)
    Callable<String> mockCallable;

    @Mock
    RingBuffer<RetryStatus> mockRingBuffer;


    @Test
    public void testTrivialExecution() throws Exception {
        final int maxTries = 1;
        when(mockCallable.call())
                .thenReturn(expectedResult);

        Optional<String> result = Retrier.of(
                mockCallable,
                mockRingBuffer,
                "test",
                0L)
                .retry(maxTries);


        assertThat("Trivial task cannot be executed with result", result.isPresent());
        assertThat("Result value is incorrect", result.get(), is(expectedResult));
        verify(mockCallable, times(1)).call();
        verify(mockRingBuffer, times(1)).add(eq(RetryStatus.STARTED));
    }

    @Test
    public void testRepeatedExecution() throws Exception {
        final int maxTries = 10;
        final int failures = 6;

        OngoingStubbing<String> stub = when(mockCallable.call());
        for (int i = 0; i < failures; ++i) {
            stub = stub.thenThrow(new Exception("Failed, " + i));
        }
        stub.thenReturn(expectedResult);

        Optional<String> result = Retrier.of(
                mockCallable,
                mockRingBuffer,
                "test",
                0L)
                .retry(maxTries);


        assertThat("Repeated task cannot be executed with result", result.isPresent());
        assertThat("Result value is incorrect", result.get(), is("result"));
        verify(mockCallable, atMost(maxTries)).call();
        verify(mockRingBuffer, times(1)).add(RetryStatus.STARTED);
        verify(mockRingBuffer, times(failures)).add(eq(RetryStatus.FAILED));
    }

    @Test
    public void testInterruptedExecution() throws Exception {
        final int maxTries = 10;

        when(mockCallable.call())
                .thenThrow(new Exception("Failed"));

        Thread retrierThread = new Thread(
            () -> Retrier.of(
                    mockCallable,
                    mockRingBuffer,
                    "test",
                    2000L)
                    .retry(maxTries)
        );

        try {
            retrierThread.start();
            Thread.sleep(500);
            retrierThread.interrupt();
            retrierThread.join();
        } catch (InterruptedException e) {
            // nop
        }

        verify(mockCallable, times(0)).call();
        verify(mockRingBuffer, times(1)).add(eq(RetryStatus.INTERRUPTED));
    }

}